// to make the proc_macros work inside the microrm crate; needed for tests and the metaschema.
extern crate self as microrm;

pub mod db;
mod query;
pub mod schema;

pub mod prelude {
    pub use crate::query::{AssocInterface, Queryable};
    // pub use crate::schema::entity::Entity;
    pub use crate::schema::{AssocMap, Database, IDMap};
    pub use microrm_macros::{Database, Entity};
}

// ----------------------------------------------------------------------
// Generically-useful database types
// ----------------------------------------------------------------------

#[derive(Debug)]
pub enum Error {
    EmptyResult,
    UnknownValue(String),
    IncompatibleSchema,
    InternalError(&'static str),
    EncodingError(std::str::Utf8Error),
    LogicError(&'static str),
    Sqlite {
        code: i32,
        msg: String,
        sql: Option<String>,
    },
    JSON(serde_json::Error),
    LockError(String),
}

impl<T> From<std::sync::PoisonError<T>> for Error {
    fn from(value: std::sync::PoisonError<T>) -> Self {
        Self::LockError(value.to_string())
    }
}

impl From<std::ffi::NulError> for Error {
    fn from(_value: std::ffi::NulError) -> Self {
        Self::InternalError("NULL pointer encountered in unexpected location")
    }
}

impl From<std::str::Utf8Error> for Error {
    fn from(value: std::str::Utf8Error) -> Self {
        Self::EncodingError(value)
    }
}

pub type DBResult<T> = Result<T, Error>;
