use super::{Datum, DatumList, DatumListRef, DatumVisitor};

impl DatumList for () {
    type Ref<'a> = &'a ();
    fn accept(&self, _: &mut impl DatumVisitor) {}
}

impl<'l> DatumListRef for &'l () {
    fn accept(&self, _: &mut impl DatumVisitor) {}
}

impl<T: Datum> DatumList for T {
    type Ref<'a> = &'a T where Self: 'a;

    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(self);
    }
}

impl<'l, T: Datum> DatumListRef for &'l T {
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(self);
    }
}

impl<T0: Datum> DatumList for (T0,) {
    type Ref<'a> = (&'a T0,) where Self: 'a;

    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
    }
}

impl<'a, T0: Datum> DatumListRef for (&'a T0,) {
    fn accept(&self, visitor: &mut impl DatumVisitor) {
        visitor.visit(&self.0);
    }
}

macro_rules! datum_list {
    ($($ty:ident : $n:tt),+) => {
        impl<$($ty: Datum),*> DatumList for ($($ty),*) {
            type Ref<'a> = ($(&'a $ty),*) where Self: 'a;
            fn accept(&self, visitor: &mut impl DatumVisitor) {
                $(visitor.visit(&self. $n));*
            }
        }

        impl<'l, $($ty: Datum),*> DatumListRef for ($(&'l $ty),*) {
            fn accept(&self, visitor: &mut impl DatumVisitor) {
                $(visitor.visit(&self. $n));*
            }
        }
    }
}

datum_list!(T0: 0, T1: 1);
datum_list!(T0: 0, T1: 1, T2: 2);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4, T5: 5);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4, T5: 5, T6: 6);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4, T5: 5, T6: 6, T7: 7);
datum_list!(T0: 0, T1: 1, T2: 2, T3: 3, T4: 4, T5: 5, T6: 6, T7: 7, T8: 8);
